/*
 * designed for tinyUSBboard Rev.3 (http://matrixstorm.com/avr/tinyusbboard/)
 * 
 * by Stephan Baerwolf, matrixstorm@gmx.de
 * 
 * NOT FOR COMMERCIAL USE
 * 
 */

#include "light_ws2812.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <avr/cpufunc.h>

#include "usbdrv.h"
#include "oddebug.h"
#include <util/delay.h>

#include <stdint.h>

/* USB Request-Commands */
#define PSCMD_CALL_ECHO		0
#define PSCMD_CALL_GETSTATUS	1
#define PSCMD_CALL_SETSTATUS	2
#define PSCMD_CALL_UNKNOWN	0xff

/* USB Response-Codes */
#define PSCMD_RETN_OKAY		(0x0)
#define PSCMD_RETN_BUSY		(0x1)
#define PSCMD_RETN_UNKNOW	(0xff)



#define	eepage_to_use	(0x1f0)
static const EEMEM uint8_t eeprom[512];

static volatile uint8_t  PSCMD_CALL_last = PSCMD_CALL_UNKNOWN;
static volatile uint16_t color[3];


void update_color(void) {
  int i;
  uint8_t __sreg;
  uint8_t buf[3];
  /*R*/ buf[1]=color[0]>>8;
  /*G*/ buf[0]=color[1]>>8;
  /*B*/ buf[2]=color[2]>>8;

#if 0
  OCR1AL = buf[1];
  OCR1BL = buf[0];
  OCR2   = buf[2];
#else
  OCR1AL = ~buf[1];
  OCR1BL = ~buf[0];
  OCR2   = ~buf[2];
#endif
  
  for (i=0;i<30;i++) {
    __sreg=SREG;
    cli();
    ws2812_sendarray(buf, 3);
    if (__sreg & _BV(SREG_I)) sei();
    _NOP();
    _NOP();
    _NOP();
    _NOP();
    _NOP();
    _NOP();
    _NOP();
    _NOP();
  }
}

static volatile uint8_t clkdivider;
ISR(TIMER0_OVF_vect) {
  clkdivider++;
  if (!clkdivider) {
    // about every 4 secnds
    uint8_t eepage[4];

    eepage[0]=color[0]>>8;
    eepage[1]=color[1]>>8;
    eepage[2]=color[2]>>8;
    eepage[3]=0;

    eeprom_update_block(eepage, (void*)&eeprom[eepage_to_use], sizeof(eepage));
  }
}



uchar usbFunctionWrite(uchar *data, uchar len) {
  switch (PSCMD_CALL_last) {
    case PSCMD_CALL_SETSTATUS: {
      if (len >= 6) {
	uint8_t i;
	for (i=0;i<3;i++) {
	  color[i] = ((uint16_t)data[(i<<1)+1]<<0x8) | (uint16_t)data[(i<<1)+0];
	}
	TIMSK &= ~(_BV(TOIE0));
	update_color();
	clkdivider=0;
	TCNT0  =  0;
	TIMSK |= _BV(TOIE0);
      }
      return 1;
    }

    default: {
      return 1;
    }
  }
}

USB_PUBLIC usbMsgLen_t usbFunctionSetup(uchar data[8]) {
usbRequest_t    *rq = (void *)data;
static uchar    replyBuf[8];

    usbMsgPtr = (usbMsgPtr_t)replyBuf;
    
    replyBuf[0] = rq->bRequest;
    PSCMD_CALL_last = rq->bRequest;
    switch (rq->bRequest) {
      
      case PSCMD_CALL_ECHO: {
	replyBuf[1] = PSCMD_RETN_OKAY;
        replyBuf[2] = rq->wValue.bytes[0];
        replyBuf[3] = rq->wValue.bytes[1];
	return 4;
      }

      case PSCMD_CALL_SETSTATUS: {
	return USB_NO_MSG;
      }

      case PSCMD_CALL_GETSTATUS: {
	uint8_t i;
	replyBuf[1] = PSCMD_RETN_OKAY;
	for (i=0;i<3;i++) {
	  replyBuf[(i<<1)+2]=color[i]&0xff;
	  replyBuf[(i<<1)+3]=color[i]>>0x8;
	}
	return 8;
      }

      default: {
	replyBuf[1] = PSCMD_RETN_UNKNOW;
	return 2;
      }
    }
    
    return 0;
}

int main(void)
{
    uint16_t pressed_helper = 0xffff;

    wdt_enable(WDTO_1S);
    usbDeviceDisconnect();
    wdt_reset();

    /* time0 for continuosly update eeprom with last color */
    TCCR0  = _BV(CS02) | _BV(CS00); // maximum prescale: update rate will be still large enough
    TCNT0  =  0;
    TIMSK |= _BV(TOIE0);


    /* ** Fast PWM for brightness ** */    
    color[0]=((uint16_t)eeprom_read_byte((void*)&eeprom[eepage_to_use+0])<<8)|0x80;
    color[1]=((uint16_t)eeprom_read_byte((void*)&eeprom[eepage_to_use+1])<<8)|0x80;
    color[2]=((uint16_t)eeprom_read_byte((void*)&eeprom[eepage_to_use+2])<<8)|0x80;

    DDRD |= _BV(DDD3);
    update_color();

    // PWM TIMER1 (R&G)
    TCCR1A = 0b10100001;
    TCCR1B = 0b00001001;
    
    // PWM TIMER 2 (B)
    TCCR2 = 0b01101001;

    // all PWM ports as colorchannels OUT
    DDRB |= _BV(DDB1) | _BV(DDB2) | _BV(DDB3);
    DDRD &= ~(_BV(DDD6));

    /* USB activation stuff */
    _delay_ms(500);
    wdt_reset();

    usbDeviceConnect();
    usbInit();

    /* activate all interrupts */
    sei();
    
    for(;;){    /* main event loop */
        wdt_reset();
	usbPoll();
	if (PIND & _BV(PD6)) {
	  // nothing pressed
	  if (pressed_helper) pressed_helper--;
	} else {
	  // pressed
	  if (!pressed_helper) {
	    // debounce timeout passed
	    
	  }
	}
    }
}
